//*****************************************************************
//* PCX to Sega (CD) tile converter. (Paletted)
//*****************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>

#define PAL_SIZE		16
#define TILEDIMENSION	8
#define	BANKSIZE		(64 * 1024)

typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;
typedef BYTE			BOOL;

#ifndef	TRUE
#define	TRUE	1
#endif
#ifndef	FALSE
#define	FALSE	0
#endif

DWORD	Width;
DWORD	Height;
char	FilePal[32];
BYTE	*Bank;
WORD	Palette[PAL_SIZE];
DWORD	Offset		= 0;
DWORD	Palettes	= 1;
BYTE	PcxBuffer[1024 * 1024];
BYTE	PcxPalette[768];

struct PCXHeader
{
	BYTE ID;
	BYTE Version;
	BYTE RLE;
	BYTE BPP;
	WORD StartX;
	WORD StartY;
	WORD EndX;
	WORD EndY;
	WORD HRes;
	WORD VRes;
	BYTE Palette[48];
	BYTE Reserved1;
	BYTE BitPlanes;
	WORD BytesPerLine;
	WORD PaletteType;
	WORD HSize;
	WORD VSize;
	BYTE Reserved2[54];
};

void Usage();
long Rip(char *Filename);
void WriteTiles(char *Filename);
BOOL LoadPCX(char *File, BYTE *Buffer, BYTE *Palette);
void GetTile(BYTE *Tile, DWORD X, DWORD Y);

void main(int argc, char *argv[])
{
	printf("PCX to Sega (CD) tile converter.\n");

	if (argc < 4)
	{
		Usage();
	}

	if (6 == argc)
	{
		Palettes	= (DWORD)atol(argv[5]);
	}

	memset(Palette, 0, PAL_SIZE * 2);
    
	Offset	= atol(argv[1]);

	Bank	= new BYTE[BANKSIZE];

	if (NULL == Bank)
	{
		printf("Unable to allocate memory.\n");
			
		return;
	}

	memset(Bank, 0, BANKSIZE);

	if (Offset > 0)
	{
		FILE	*Handle = fopen(argv[3], "rb");

		if (Handle != NULL)
		{
			fseek(Handle, 0, SEEK_END);

			DWORD	Length	= ftell(Handle);
			rewind(Handle);

			fread(Bank, 1, Length, Handle);
			fclose(Handle);
		}
	}

	if (argc > 4)
	{
		strcpy(FilePal, argv[4]);

		if (strlen(FilePal) <= 0)
		{
			printf("Palette will not be saved.\n");
		}
	}

	else
	{
		memset(FilePal, 0, sizeof(FilePal));
	}

	if (-1 == Rip(argv[2]))
	{
		exit(1);
	}

	WriteTiles(argv[3]);

	printf("Next Offset = %d\n", Offset); 
}

void Usage()
{
	printf("Command syntax is: PcxSCD Offset Infile.pcx Outfile.bin [Palette] [Palettes]\n");

	exit(1);
}

long Rip(char *Filename)
{
	BYTE	*Buffer	= Bank + Offset;
	BYTE	Tile[TILEDIMENSION * TILEDIMENSION];

	LoadPCX(Filename, PcxBuffer, PcxPalette);

	if ((Width % TILEDIMENSION) != 0 && (Height % TILEDIMENSION) != 0)
	{
		printf("Picture must be a multiple of %d.\n", TILEDIMENSION);
        
		return	-1;
	}

	for (DWORD yloop = 0; yloop < (Height / TILEDIMENSION); yloop++)
	{
		for (DWORD xloop = 0; xloop < (Width / TILEDIMENSION); xloop++)
		{
			DWORD	x	= xloop * TILEDIMENSION;
			DWORD	y	= yloop * TILEDIMENSION;

			GetTile(Tile, x, y);

			for (DWORD TileY = 0; TileY < TILEDIMENSION; TileY++)
			{
				for (DWORD TileX = 0; TileX < TILEDIMENSION; TileX	+= 2)
				{
					BYTE Pixel1	= (Tile[TileY * TILEDIMENSION + TileX]) & 0x0F;
					BYTE Pixel2	= (Tile[TileY * TILEDIMENSION + TileX + 1]) & 0x0F;
				
					*Buffer	= (Pixel1 << 4) | Pixel2;
				
					Buffer++;
					Offset++;
				}
			}
		}
	}
    
	return	0;
}

void WriteTiles(char *Filename)
{
	for (DWORD Loop = 0; Loop < PAL_SIZE; Loop++)
	{
		BYTE	Red    = ((PcxPalette[Loop * 3]) >> 2);
		BYTE	Green  = ((PcxPalette[Loop * 3 + 1]) >> 2);
		BYTE	Blue   = ((PcxPalette[Loop * 3 + 2]) >> 2);

		Red		&= 0x0F;
		Green	&= 0x0F;
		Blue	&= 0x0F;

		WORD	Value	= ((WORD)Blue << 0) | ((WORD)Green << 12) | ((WORD)Red << 8);
		Palette[Loop]	= Value;	
	}

	FILE	*Handle;

	if (strlen(FilePal) > 0)
	{
		Handle	= fopen(FilePal, "wb");

		if (NULL == Handle)
		{
			return;
		}
		
		fwrite(Palette, 2, (Palettes * 32) / sizeof(WORD), Handle);
		fclose(Handle);
	}

	Handle	= fopen(Filename, "wb");

	if (NULL == Handle)
	{
		return;
	}

	fwrite(Bank, 1, Offset, Handle);
	fclose(Handle);
}

BOOL LoadPCX(char *File, BYTE *Buffer, BYTE *Palette)
{
	PCXHeader   PcxHeader;
	DWORD		Position;
	DWORD		Length;

	FILE	*Handle	= fopen(File, "rb");

	if (NULL == Handle)
	{
		printf("Unable to open %s.\n", File);

		return	FALSE;
	}

	Position	= ftell(Handle);
	fseek(Handle, 0, SEEK_END);
	Length		= ftell(Handle);
	rewind(Handle);

	fread(&PcxHeader, 1, sizeof(PcxHeader), Handle);

	if (PcxHeader.ID != 0x0A || PcxHeader.Version != 5 || PcxHeader.RLE != 1 
		|| (8 == PcxHeader.BPP && PcxHeader.BitPlanes != 1))
	{
		return	FALSE;
	}

	Width   = PcxHeader.EndX - PcxHeader.StartX + 1;
	Height  = PcxHeader.EndY - PcxHeader.StartY + 1;

	if (1 == PcxHeader.BPP && 4 == PcxHeader.BitPlanes)
	{
		 memset(Palette, 0, 768);

		for (DWORD loop = 0; loop < 48; loop++)
		{
			Palette[loop]	= PcxHeader.Palette[loop] >> 2;
		}
	}

	else if (8 == PcxHeader.BPP && 1 == PcxHeader.BitPlanes)
	{
		fseek(Handle, Position + Length - 768, SEEK_SET);

		for (DWORD loop = 0; loop < 768; loop++)
		{
			Palette[loop]	= ((BYTE)getc(Handle)) >> 2;
		}
	}

	else
	{
		return	FALSE;
	}

	fseek(Handle, Position + sizeof(PCXHeader), SEEK_SET);

	DWORD	TmpHeight	= Height;

	while (TmpHeight-- > 0)
	{
		BYTE	Pixel;
		BYTE	*Output;

		Output	= Buffer;

		memset(Buffer, 0, Width);

		for (DWORD Planes = 0; Planes < PcxHeader.BitPlanes; Planes++)
		{
			DWORD	Done	= 0;
			Buffer	= Output;

			do
			{
				Pixel	= (BYTE)getc(Handle);

				if ((Pixel & 0xC0) != 0xC0)
				{
					if (1 == PcxHeader.BPP)
					{
						for (DWORD Bits = 7; Bits >= 0; Bits--)
						{
							*Buffer++	|= ((Pixel >> Bits) & 1) << Planes;
						}

						Done	+= 8;
					}

					else
					{
						*Buffer++	= Pixel;
						Done++;
					}
				}

				else
				{
					BYTE	Run	= (BYTE)getc(Handle);
					Pixel	&= ~0xC0;

					while (Pixel > 0 && Done < Width)
					{
						if (1 == PcxHeader.BPP)
						{
							for (DWORD Bits = 7; Bits >= 0; Bits--)
							{
								*Buffer++	|= ((Run >> Bits) & 1) << Planes;
							}

							Done	+= 8;
						}

						else
						{
							*Buffer++	= Run;
							Done++;
						}

						Pixel--;
					}
				}
			}
			
			while (Done < Width);
		}
	}

	if (Handle != NULL)
	{
		fclose(Handle);
	}
	
	return	TRUE;
}

void GetTile(BYTE *Tile, DWORD X, DWORD Y)
{
	BYTE *Buffer	= PcxBuffer;

	Buffer += Y * Width + X;

	for (DWORD OuterLoop = 0; OuterLoop < TILEDIMENSION; OuterLoop++)
	{
		for (DWORD InnerLoop = 0; InnerLoop < TILEDIMENSION; InnerLoop++)
		{
			*Tile	= *Buffer;
			Tile++;
			Buffer++;
		}

		Buffer += Width - TILEDIMENSION;
	}
}
