//*****************************************************************
//* TUME Room to Sega (CD) map converter.
//*****************************************************************

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <string.h>

typedef unsigned char	BYTE;
typedef unsigned short	WORD;
typedef unsigned long	DWORD;
typedef BYTE			BOOL;

#ifndef TRUE
#define TRUE	1
#endif
#ifndef FALSE
#define FALSE	0
#endif

#define	TILEDIMENSION	8
#define MAXTILES		4096

char	FileIn[32];
char	FileOut[32];
char	PcxFile[32];
WORD	BankIn[1024 * 1024];
WORD	BankOut[1024 * 1024];
DWORD	Offset		= 0;
DWORD	Priority	= 0;
DWORD	Length;
DWORD	Width;
DWORD	Height;
DWORD	MapWidth;
DWORD	MapHeight;
BYTE	PcxBuffer[1024 * 1024];
BYTE	PcxPalette[768];
BYTE	TilePalettes[MAXTILES];
WORD	PaletteOverride	= 0xFFFF;

struct PCXHeader
{
	BYTE ID;
	BYTE Version;
	BYTE RLE;
	BYTE BPP;
	WORD StartX;
	WORD StartY;
	WORD EndX;
	WORD EndY;
	WORD HRes;
	WORD VRes;
	BYTE Palette[48];
	BYTE Reserved1;
	BYTE BitPlanes;
	WORD BytesPerLine;
	WORD PaletteType;
	WORD HSize;
	WORD VSize;
	BYTE Reserved2[54];
};

void Usage();
void ConvertMap();
BOOL LoadPCX(char *File, BYTE *Buffer, BYTE *Palette);
void GetTile(BYTE *Tile, DWORD X, DWORD Y);
DWORD CheckTile(BYTE *Tile);

void main(int argc, char *argv[])
{
	printf("TUME Room to Sega (CD) map converter.\n");

    if (argc < 7)
    {
        Usage();
    }

	if (argc > 7)
	{
		Priority	= atol(argv[7]);
	}

	if (9 == argc)
	{
		PaletteOverride	= (WORD)atol(argv[8]);
	}

	memset(BankIn, 0, sizeof(BankIn));
	memset(BankOut, 0, sizeof(BankOut));
    
	Offset		= atoi(argv[1]);
	MapWidth	= atoi(argv[2]);
	MapHeight	= atoi(argv[3]);

    strcpy(FileIn, argv[4]);
    strcpy(FileOut, argv[5]);
    strcpy(PcxFile, argv[6]);

    if ("" == FileIn || "" == FileOut || "" == PcxFile)
    {
        Usage();
    }

    FILE *handle = fopen(FileIn, "rb");

	if (handle != NULL)
    {
		fseek(handle, 0, SEEK_END);
        Length	= ftell(handle) / 2;
		rewind(handle);

		if (Length > sizeof(BankIn) / 2)
		{
			Length = sizeof(BankIn) / 2;
		}

        fread(BankIn, 2, Length, handle);
		fclose(handle);
    }

    ConvertMap();
}

void Usage()
{
    printf("Command syntax is: RoomSCD Offset MapWidth MapHeight Infile Outfile Tiles.pcx [Priority] [Palette]\n");

    exit(1);
}

void ConvertMap()
{
	BYTE	Tile[TILEDIMENSION * TILEDIMENSION];

	LoadPCX(PcxFile, PcxBuffer, PcxPalette);

    DWORD	Tiles	= 0;

	for (DWORD yloop = 0; yloop < (Height / TILEDIMENSION); yloop++)
    {
        for (DWORD xloop = 0; xloop < (Width / TILEDIMENSION); xloop++)
        {
			GetTile(Tile, xloop * TILEDIMENSION, yloop * TILEDIMENSION);
			
			if ((xloop + yloop) != 0 && FALSE == CheckTile(Tile))
			{
				continue;
			}
			
			if (0xFFFF == PaletteOverride)
			{
				TilePalettes[Tiles]	= Tile[0] >> 4;
			}

			else
			{
				TilePalettes[Tiles]	= (BYTE)PaletteOverride;
			}

			Tiles++;
		}
	}

	DWORD	Counter	= 0;

	for (DWORD OuterLoop = 0; OuterLoop < MapHeight; OuterLoop++)
	{
		for (DWORD InnerLoop = 0; InnerLoop < MapWidth; InnerLoop++)
		{
			WORD Map	= BankIn[OuterLoop * MapWidth + InnerLoop] + (WORD)Offset;
			
//			if (Map > 0x07FF)
//			{
//				BankOut[Counter]	= 0;
//			}

			BankOut[Counter]		= ((Map & 0x0700) >> 8) | ((Map & 0xFF) << 8);
			BankOut[Counter]		|= TilePalettes[(Map & 0x07FF) - Offset] << 5;

			if (Priority != 0)
			{
				BankOut[Counter]	|= 1 << 7;
			}

			Counter++;
		}
	
		Counter	+= 64 - MapWidth;
	}
			
	FILE *handle = fopen(FileOut, "wb");

	if (NULL == handle)
	{
		return;
	}

	fwrite(BankOut, 2, MapHeight * 64, handle);
	fclose(handle);
}

BOOL LoadPCX(char *File, BYTE *Buffer, BYTE *Palette)
{
	FILE		*handle;
	PCXHeader   PcxHeader;
	DWORD		Position;
	DWORD		Length;

	handle	= fopen(File, "rb");

	if (NULL == handle)
	{
		return	FALSE;
	}

	Position	= ftell(handle);
	fseek(handle, 0, SEEK_END);
	Length		= ftell(handle);
	rewind(handle);

	fread(&PcxHeader, 1, sizeof(PcxHeader), handle);

	if (PcxHeader.ID != 0x0A || PcxHeader.Version != 5 || PcxHeader.RLE != 1 
		|| (8 == PcxHeader.BPP && PcxHeader.BitPlanes != 1))
	{
		return	FALSE;
	}

	Width   = PcxHeader.EndX - PcxHeader.StartX + 1;
	Height  = PcxHeader.EndY - PcxHeader.StartY + 1;

	if (1 == PcxHeader.BPP && 4 == PcxHeader.BitPlanes)
	{
		 memset(Palette, 0, 768);

		for (DWORD loop = 0; loop < 48; loop++)
		{
			Palette[loop]	= PcxHeader.Palette[loop] >> 2;
		}
	}

	else if (8 == PcxHeader.BPP && 1 == PcxHeader.BitPlanes)
	{
		fseek(handle, Position + Length - 768, SEEK_SET);

		for (DWORD loop = 0; loop < 48; loop++)
		{
			Palette[loop]	= ((BYTE)getc(handle)) >> 2;
		}
	}

	else
	{
		return	FALSE;
	}

	fseek(handle, Position + sizeof(PCXHeader), SEEK_SET);

	DWORD	TmpHeight	= Height;

	while (TmpHeight-- > 0)
	{
		BYTE	Pixel;
		BYTE	*Output;

		Output	= Buffer;

		memset(Buffer, 0, Width);

		for (DWORD Planes = 0; Planes < PcxHeader.BitPlanes; Planes++)
		{
			DWORD	Done	= 0;
			Buffer	= Output;

			do
			{
				Pixel	= (BYTE)getc(handle);

				if ((Pixel & 0xC0) != 0xC0)
				{
					if (1 == PcxHeader.BPP)
					{
						for (DWORD Bits = 7; Bits >= 0; Bits--)
						{
							*Buffer++	|= ((Pixel >> Bits) & 1) << Planes;
						}

						Done	+= 8;
					}

					else
					{
						*Buffer++	= Pixel;
						Done++;
					}
				}

				else
				{
					BYTE	Run	= (BYTE)getc(handle);
					Pixel	&= ~0xC0;

					while (Pixel > 0 && Done < Width)
					{
						if (1 == PcxHeader.BPP)
						{
							for (DWORD Bits = 7; Bits >= 0; Bits--)
							{
								*Buffer++	|= ((Run >> Bits) & 1) << Planes;
							}

							Done	+= 8;
						}

						else
						{
							*Buffer++	= Run;
							Done++;
						}

						Pixel--;
					}
				}
			}
			
			while (Done < Width);
		}
	}

	if (handle != NULL)
	{
		fclose(handle);
	}
	
	return	TRUE;
}

void GetTile(BYTE *Tile, DWORD X, DWORD Y)
{
	BYTE *Buffer	= PcxBuffer;

	Buffer += Y * Width + X;

	for (DWORD outerloop = 0; outerloop < TILEDIMENSION; outerloop++)
	{
		for (DWORD innerloop = 0; innerloop < TILEDIMENSION; innerloop++)
		{
			*Tile++	= *Buffer++;
		}

		Buffer += Width - TILEDIMENSION;
	}
}

DWORD CheckTile(BYTE *Tile)
{
	for (DWORD Loop = 0; Loop < TILEDIMENSION * TILEDIMENSION; Loop++)
	{
		if (Tile[Loop] != 0)
		{
			return	TRUE;
		}
	}
	
	return	FALSE;
}

